module L = Llvm
module A = Ast 
open Sast 

module StringMap = Map.Make(String)

let translate (globals, functions) = 
  let context = L.global_context() in 
  let the_module = L.create_module context "nodable" in 

  let i32_t = L.i32_type context
  and i8_t = L.i8_type context
  and i1_t       = L.i1_type     context (* bool *)
  and float_t    = L.double_type context
  and str_t = L.pointer_type (L.i8_type context)
  and obj_ptr_t  = L.pointer_type (L.i8_type context)
  and void_ptr_t = L.pointer_type (L.i8_type context)
  and void_t = L.void_type context in 

  (* Return the LLVM type for a nodable type *)
  let ltype_of_typ = function 
    A.Int -> i32_t 
  (*| A.Bool  -> i1_t*)
  | A.Void -> void_t 
  | A.Bool -> i1_t
  | A.Float -> float_t
  | A.String -> str_t
  | A.List(_)  -> obj_ptr_t
  | A.Node(_) -> obj_ptr_t 
  | _ -> raise (Failure "not implemented")
  in 

  (* Create a map of global variables after creating each *)
  let global_vars : L.llvalue StringMap.t =
    let global_var m (t, n) = 
      let init = match t with
          A.Float -> L.const_float (ltype_of_typ t) 0.0
        | _ -> L.const_int (ltype_of_typ t) 0
      in StringMap.add n (L.define_global n init the_module) m in
    List.fold_left global_var StringMap.empty globals in 

    (* string functions*)

     (* NODE FUNCTIONS *)

    let init_t : L.lltype = 
      L.var_arg_function_type obj_ptr_t [| void_ptr_t |] in

  let init_node : L.llvalue = 
      L.declare_function "init_node" init_t the_module in

  let get_t : L.lltype = 
    L.var_arg_function_type void_ptr_t [|obj_ptr_t|] in
  let get_data : L.llvalue = 
      L.declare_function "get_data" get_t the_module in


  let get_left_t : L.lltype = 
    L.var_arg_function_type void_ptr_t [|obj_ptr_t|] in
  let get_left : L.llvalue = 
      L.declare_function "get_left" get_left_t the_module in

  let get_right_t : L.lltype = 
      L.var_arg_function_type obj_ptr_t [|obj_ptr_t|] in
  let get_right : L.llvalue = 
      L.declare_function "get_right" get_right_t the_module in


let add_left_t : L.lltype = 
  L.var_arg_function_type obj_ptr_t [|obj_ptr_t|] in
let add_left : L.llvalue = 
    L.declare_function "add_left" add_left_t the_module in

let add_right_t : L.lltype = 
    L.var_arg_function_type obj_ptr_t [|obj_ptr_t|] in
let add_right : L.llvalue = 
    L.declare_function "add_right" add_right_t the_module in
  
    (* list functions *)

  let list_init_t : L.lltype = 
      L.var_arg_function_type obj_ptr_t [||] in
  let list_init : L.llvalue = 
      L.declare_function "list_init" list_init_t the_module in

  let get_elem_t : L.lltype = 
    L.var_arg_function_type void_ptr_t [|i32_t ; obj_ptr_t|] in
  let get_elem : L.llvalue = 
    L.declare_function "get_elem" get_elem_t the_module in

  let add_elem_t : L.lltype = 
      L.var_arg_function_type void_ptr_t [|obj_ptr_t; void_ptr_t |] in
  let add_elem : L.llvalue = 
      L.declare_function "add_elem" add_elem_t the_module in

  let append : L.lltype = 
    L.var_arg_function_type void_ptr_t [|obj_ptr_t; void_ptr_t |] in
  let append : L.llvalue = 
      L.declare_function "append" append the_module in

  let update_elem_t : L.lltype = 
    L.var_arg_function_type obj_ptr_t [|void_ptr_t ; obj_ptr_t ; i32_t|] in
  let update_elem : L.llvalue = 
      L.declare_function "update_elem" update_elem_t the_module in

  let size_t : L.lltype = 
    L.var_arg_function_type i32_t [|obj_ptr_t|] in
  let size : L.llvalue = 
      L.declare_function "size" size_t the_module in

  let printf_t : L.lltype = 
    L.var_arg_function_type i32_t [|L.pointer_type i8_t|] in 
  let printf_func : L.llvalue = 
    L.declare_function "printf" printf_t the_module in

  let function_decls: (L.llvalue * sfunc_decl) StringMap.t = 
    let function_decl m fdecl = 
      let name = fdecl.sfname
      and formal_types = 
    Array.of_list (List.map(fun(t, _) -> ltype_of_typ t) fdecl.sformals)
      in let ftype = 
        L.function_type(ltype_of_typ fdecl.styp) formal_types in 
      StringMap.add name (L.define_function name ftype the_module, fdecl) m in 
    List.fold_left function_decl StringMap.empty functions in 

    let build_function_body fdecl = 
      let (the_function, _) = 
        StringMap.find fdecl.sfname function_decls in 
      let builder = 
        L.builder_at_end context (L.entry_block the_function) in 
      let int_format_str = 
        L.build_global_stringptr "%d\n" "fmt" builder
      and float_format_str = L.build_global_stringptr "%g\n" "fmt" builder
      and str_format_str = L.build_global_stringptr "%s\n" "fmt" builder in

   (* Construct the function's "locals": formal arguments and locally
       declared variables.  Allocate each on the stack, initialize their
       value, if appropriate, and remember their values in the "locals" map *)
       let local_vars =
        let add_formal m (t, n) p = 
          L.set_value_name n p;
    let local = L.build_alloca (ltype_of_typ t) n builder in
          ignore (L.build_store p local builder);
    StringMap.add n local m 

      (* Allocate space for any locally declared variables and add the
       * resulting registers to our map *)
       and add_local m (t, n) =
       let local_var = L.build_alloca (ltype_of_typ t) n builder
       in StringMap.add n local_var m 
           in
     
           let formals = List.fold_left2 add_formal StringMap.empty fdecl.sformals
               (Array.to_list (L.params the_function)) in
            let rec declare locals = function
               [] -> locals
             | hd::tl -> declare (match hd with
                  SDeclare (t, id, _) -> (t, id) :: locals
                | _ -> locals) tl
           in

          let declarations = declare [] fdecl.sbody in
            List.fold_left add_local formals declarations
          in

         (* Return the value for a variable or formal argument.
            Check local names first, then global names *)
         let lookup n = try StringMap.find n local_vars
                        with Not_found -> StringMap.find n global_vars
         in

    (* Construct code for an expression; return its value *)
    let rec expr builder ((_, e):sexpr) = match e with 
        SLiteral i -> L.const_int i32_t i 
        | SBoolLit b  -> L.const_int i1_t (if b then 1 else 0)
        | SFliteral l -> L.const_float_of_string float_t l
        | SStrLit l -> L.build_global_stringptr l "str" builder
        | SNoexpr -> L.const_int i32_t 0
        | SId s       -> L.build_load (lookup s) s builder
        | SAssign (s, e) -> let e' = expr builder e in
        ignore(L.build_store e' (lookup s) builder); e'
        | SListAccess (l, e) ->
          let l' = expr builder l in
          let e' = expr builder e in 
          (match fst l with 
             A.List t -> 
              let d_ptr = L.build_call get_elem [|e'; l'|] "get_elem" builder in  
              match t with 
                  A.List _ | A.Node _ -> d_ptr
                | _ -> 
                  let dt_ptr = L.pointer_type (ltype_of_typ t) in
                  let d_ptr = L.build_bitcast d_ptr dt_ptr "data" builder in
                  L.build_load d_ptr "data" builder)

        | SAttr (o, p) ->
          let tobj = fst o in
          let o' = expr builder o in
          (match (tobj, p) with
          | (A.Node t, "data") ->
              let dt_ptr = L.pointer_type (ltype_of_typ t) in
              let d_ptr = L.build_call get_data [|o'|] "data" builder in  
              let d_ptr = L.build_bitcast d_ptr dt_ptr "data" builder in
              L.build_load d_ptr "data" builder
          | (A.Node t, "left") ->
            L.build_call get_left [|o'|] "left" builder
          | (A.Node t, "right") ->
              L.build_call get_right [|o'|] "right" builder
          | (_, _) -> raise (Failure "no such property"))
      
          | SNodeLit (t, v) ->
          let data_value = expr builder (t, v) in 
          let data = L.build_malloc (ltype_of_typ t) "data_malloc" builder in
            ignore ( L.build_store data_value data builder);
          let data = L.build_bitcast data void_ptr_t "data_bitcast" builder in
          let node = L.build_call init_node [|data|] "init_node" builder in node

        | SListLit li ->
          let rec create_list mylist = (function 
          [] -> mylist
          |h :: tail ->
          let (anyt,_) = h in
          let d_ptr = (match anyt with
            A.List _ | A.Node _ -> expr builder h 
            | _  -> let elem = L.build_malloc (ltype_of_typ anyt) "data" builder in
              let elem_value = expr builder h in  
              ignore (L.build_store elem_value elem builder); elem) in 
          let elem = L.build_bitcast d_ptr void_ptr_t "data" builder in 
          L.build_call add_elem [|mylist; elem|] "" builder;
           create_list mylist tail) in
          let mylist = L.build_call list_init [||] "list_init" builder in create_list mylist li     
        | SBinop ((A.Float,_ ) as e1, op, e2) ->
          let e1' = expr builder e1
          and e2' = expr builder e2 in
          (match op with 
              A.Add     -> L.build_fadd
            | A.Sub     -> L.build_fsub
            | A.Mult    -> L.build_fmul
            | A.Div     -> L.build_fdiv 
            | A.Mod     -> L.build_frem
            | A.Eq      -> L.build_fcmp L.Fcmp.Oeq
            | A.Neq     -> L.build_fcmp L.Fcmp.One
            | A.Less    -> L.build_fcmp L.Fcmp.Olt
            | A.Leq     -> L.build_fcmp L.Fcmp.Ole
            | A.Greater -> L.build_fcmp L.Fcmp.Ogt
            | A.Geq     -> L.build_fcmp L.Fcmp.Oge
            | A.And | A.Or ->
              raise (Failure "internal error: semant should have rejected and/or on float")
          ) e1' e2' "tmp" builder
      | SBinop (e1, op, e2) ->
      let e1' = expr builder e1
      and e2' = expr builder e2 in
      (match op with
        A.Add     -> L.build_add
      | A.Sub     -> L.build_sub
      | A.Mult    -> L.build_mul
      | A.Div     -> L.build_sdiv
      | A.Mod     -> L.build_srem
      | A.And     -> L.build_and
      | A.Or      -> L.build_or
      | A.Eq      -> L.build_icmp L.Icmp.Eq
      | A.Neq     -> L.build_icmp L.Icmp.Ne
      | A.Less    -> L.build_icmp L.Icmp.Slt
      | A.Leq     -> L.build_icmp L.Icmp.Sle
      | A.Greater -> L.build_icmp L.Icmp.Sgt
      | A.Geq     -> L.build_icmp L.Icmp.Sge
      ) e1' e2' "tmp" builder
        | SUnop(op, ((t, _) as e)) ->
            let e' = expr builder e in
      (match op with
        A.Neg when t = A.Float -> L.build_fneg 
      | A.Neg                  -> L.build_neg
            | A.Not                  -> L.build_not) e' "tmp" builder        
          | SCall ("print", [e])
          | SCall ("printb", [e]) -> L.build_call printf_func [| int_format_str ; (expr builder e) |]"printf" builder
          | SCall ("printf", [e]) -> L.build_call printf_func [| float_format_str ; (expr builder e) |] "printf" builder
        (*| SCall ("printf", [e]) -> L.build_call printf_func [| int_format_str ; (expr builder e) |]"printf" builder *)
          | SCall ("prints", [e]) -> L.build_call printf_func [| str_format_str ; (expr builder e) |]"printf" builder
          | SCall ("size", [e]) -> L.build_call size [|expr builder e|] "size" builder
          | SCall ("append", [e; f]) -> L.build_call append [|expr builder e; expr builder f|] "append" builder
          | SCall ("update_elem", [e;f;g]) -> L.build_call update_elem [|expr builder e; expr builder f; expr builder g|] "update_elem" builder
          | SCall ("add_left", [e; f]) -> L.build_call add_left [|expr builder e; expr builder f|] "add_left" builder
          | SCall ("add_right", [e; f]) -> L.build_call add_right [|expr builder e; expr builder f|] "add_right" builder
          | SCall ("get_left", [e]) -> L.build_call get_left [|expr builder e|] "get_left" builder
          | SCall ("get_right", [e]) -> L.build_call get_right [|expr builder e|] "get_right" builder
          | SCall (f, args) ->
          let (fdef, fdecl) = StringMap.find f function_decls in
    let llargs = List.rev (List.map (expr builder) (List.rev args)) in
    let result = (match fdecl.styp with 
                         A.Void -> ""
                       | _ -> f ^ "_result") in
          L.build_call fdef (Array.of_list llargs) result builder
     in
     
     (* LLVM insists each basic block end with exactly one "terminator" 
        instruction that transfers control.  This function runs "instr builder"
        if the current block does not already have a terminator.  Used,
        e.g., to handle the "fall off the end of the function" case. *)
      
        let add_terminal builder instr = 
        match L.block_terminator (L.insertion_block builder) with
        Some _ -> ()
        | None -> ignore (instr builder) in 


    let rec stmt builder = function
    SBlock sl -> List.fold_left stmt builder sl
    | SExpr e -> ignore(expr builder e); builder 
    | SDeclare (_, _, a) -> ignore(expr builder a); builder
    | SReturn e -> ignore(match fdecl.styp with
                                (* Special "return nothing" instr *)
                                A.Void -> L.build_ret_void builder 
                                (* Build return statement *)
                              | _ -> L.build_ret (expr builder e) builder );
                        builder
    | SIf (predicate, then_stmt, else_stmt) ->
         let bool_val = expr builder predicate in
	 let merge_bb = L.append_block context "merge" the_function in
         let build_br_merge = L.build_br merge_bb in (* partial function *)

	 let then_bb = L.append_block context "then" the_function in
	 add_terminal (stmt (L.builder_at_end context then_bb) then_stmt)
	   build_br_merge;

	 let else_bb = L.append_block context "else" the_function in
	 add_terminal (stmt (L.builder_at_end context else_bb) else_stmt)
	   build_br_merge;

	 ignore(L.build_cond_br bool_val then_bb else_bb builder);
	 L.builder_at_end context merge_bb

      | SWhile (predicate, body) ->
	  let pred_bb = L.append_block context "while" the_function in
	  ignore(L.build_br pred_bb builder);

	  let body_bb = L.append_block context "while_body" the_function in
	  add_terminal (stmt (L.builder_at_end context body_bb) body)
	    (L.build_br pred_bb);

	  let pred_builder = L.builder_at_end context pred_bb in
	  let bool_val = expr pred_builder predicate in

	  let merge_bb = L.append_block context "merge" the_function in
	  ignore(L.build_cond_br bool_val body_bb merge_bb pred_builder);
	  L.builder_at_end context merge_bb

      (* Implement for loops as while loops *)
      | SFor (e1, e2, e3, body) -> stmt builder
	    ( SBlock [SExpr e1 ; SWhile (e2, SBlock [body ; SExpr e3]) ] )
    in


    let builder = stmt builder (SBlock fdecl.sbody)

      in add_terminal builder (match fdecl.styp with
         A.Void -> L.build_ret_void
         | t -> L.build_ret (L.const_int (ltype_of_typ t) 0))
                

in List.iter build_function_body functions; 
the_module
(* EOF *)
